'''Pipeline execution.'''
import datetime
import functools
import os
import os.path
import sys

import seesaw.util
from seesaw.config import realize
from seesaw.event import Event
from seesaw.item import Item

from tornado import ioloop


class Runner(object):
    '''Executes and manages the lifetime of :class:`Pipeline` instances.'''
    def __init__(self, stop_file=None, concurrent_items=1, max_items=None,
                 keep_data=False):
        self.pipeline = None
        self.concurrent_items = concurrent_items
        self.max_items = max_items
        self.keep_data = keep_data

        self.item_count = 0
        self.active_items = set()
        self.stop_flag = False
        self.stop_file = stop_file
        self.initial_stop_file_mtime = self.stop_file_mtime()

        self.on_status = Event()
        self.on_create_item = Event()
        self.on_pipeline_start_item = Event()
        self.on_pipeline_finish_item = Event()
        self.on_finish = Event()

        if stop_file:
            ioloop.PeriodicCallback(self.check_stop_file, 5000).start()

    def set_current_pipeline(self, pipeline):
        old_pipeline = self.pipeline

        if pipeline:
            pipeline.on_start_item += self._item_starting
            pipeline.on_finish_item += self._item_finished

        self.pipeline = pipeline

        if old_pipeline:
            # stop any cancellable items in the previous pipeline
            old_pipeline.cancel_items()

    def is_active(self):
        return len(self.active_items) > 0

    def start(self):
        self.add_items()

    def stop_gracefully(self):
        print("Stopping when current tasks are completed...")
        self.stop_flag = True
        self.pipeline.cancel_items()
        self.pipeline.on_stop_requested()
        self.initial_stop_file_mtime = self.stop_file_mtime()
        self.on_status(self, "stopping")

    def keep_running(self):
        print("Keep running...")
        self.stop_flag = False
        self.pipeline.on_stop_canceled()
        self.initial_stop_file_mtime = self.stop_file_mtime()
        self.on_status(self, "running")

    def check_stop_file(self):
        if self.stop_file_changed():
            self.stop_gracefully()

    def should_stop(self):
        return self.stop_flag or self.stop_file_changed()

    def stop_file_changed(self):
        current_stop_file_mtime = self.stop_file_mtime()
        if current_stop_file_mtime:
            return self.initial_stop_file_mtime is None \
                or self.initial_stop_file_mtime < current_stop_file_mtime
        else:
            return False

    def stop_file_mtime(self):
        if self.stop_file and os.path.exists(self.stop_file):
            return os.path.getmtime(self.stop_file)
        else:
            return None

    def add_items(self):
        if self.pipeline:
            items_required = int(realize(self.concurrent_items))
            while len(self.active_items) < items_required:
                if self.max_items and self.max_items <= self.item_count:
                    return

                self.item_count += 1
                item_id = "{0}-{1}".format(
                    seesaw.util.unique_id_str(), self.item_count)
                item = Item(
                    pipeline=self.pipeline,
                    item_id=item_id,
                    item_number=self.item_count,
                    keep_data=self.keep_data
                )
                self.on_create_item(self, item)
                self.active_items.add(item)
                self.pipeline.enqueue(item)

    def _item_starting(self, pipeline, item):
        self.on_pipeline_start_item(self, pipeline, item)

    def _item_finished(self, pipeline, item):
        if item.failed:
            item.log_output("Waiting 10 seconds...")
            ioloop.IOLoop.instance().add_timeout(
                datetime.timedelta(seconds=10),
                functools.partial(
                    self._item_finished_without_delay, pipeline, item)
            )
        else:
            self._item_finished_without_delay(pipeline, item)

    def _item_finished_without_delay(self, pipeline, item):
        self.on_pipeline_finish_item(self, pipeline, item)
        self.active_items.remove(item)

        def add_more_items():
            if not self.should_stop():
                self.add_items()

            if len(self.active_items) == 0:
                self.on_finish(self)

        ioloop.IOLoop.instance().add_timeout(
            datetime.timedelta(),
            add_more_items
        )


class SimpleRunner(Runner):
    '''Executes a single class:`Pipeline` instance.'''
    def __init__(self, pipeline, stop_file=None, concurrent_items=1,
                 max_items=None, keep_data=False):
        Runner.__init__(
            self, stop_file=stop_file,
            concurrent_items=concurrent_items, max_items=max_items,
            keep_data=keep_data)

        self.set_current_pipeline(pipeline)
        self.on_create_item += self._handle_create_item
        self.on_finish += self._stop_ioloop

    def start(self):
        Runner.start(self)
        ioloop.IOLoop.instance().start()
        self.pipeline.on_cleanup()

    def _stop_ioloop(self, dummy):
        ioloop.IOLoop.instance().stop()

    def forced_stop(self):
        print("Stopping immediately...")
        # TODO perhaps the subprocesses should be killed
        ioloop.IOLoop.instance().stop()

    def _handle_create_item(self, dummy, item):
        item.on_output += self._handle_item_output

    def _handle_item_output(self, item, data):
        while True:
            try:
                try:
                    sys.stdout.write(data)
                except UnicodeError:
                    sys.stdout.write(data.encode('ascii', 'replace').decode('ascii'))
                sys.stdout.flush()
                return
            except IOError as e:
                # Ignore EINTR errors (which are spurious errors caused by signals) and retry the operation.
                # Allow other errors to propagate up the call stack as normal.
                if e.errno != os.errno.EINTR:
                    raise



import unittest

from seesaw.item import Item


class MockPipeline(object):
    pass


class ItemTest(unittest.TestCase):
    def setUp(self):
        self.item = Item(MockPipeline(), 'FakeID', 1, prepare_data_directory=False)

    def test_get_returns_none_for_undefined_keys(self):
        self.assertEquals(None, self.item.get('undefined_key'))

    def test_get_returns_property(self):
        self.item['foo'] = 'bar'

        self.assertEquals('bar', self.item.get('foo'))

    def test_property_events(self):
        non_local_dict = {}

        def my_callback(item, key, new_value, old_value):
            self.assertEquals(self.item, item)
            non_local_dict['callback_fired'] = True
            self.assertEqual('blah', key)
            self.assertEqual('blahblah', new_value)
            self.assertEqual(None, old_value)

        self.item.on_property.handle(my_callback)
        self.item['blah'] = 'blahblah'

        self.assertTrue(non_local_dict.get('callback_fired'))



import base64
import re

import tornado
import tornado.web


try:
    callable
except NameError:
    from seesaw.six import callable


class BaseWebAdminHandler(tornado.web.RequestHandler):
    def prepare(self):
        if not self.application.settings['auth_enabled']:
            return

        for pattern in self.application.settings['skip_auth']:
            if pattern.match(self.request.uri):
                return

        auth_header = self.request.headers.get('Authorization')
        if auth_header and auth_header.startswith('Basic '):
            auth_decoded = base64.b64decode(auth_header[6:].encode('ascii')).decode('ascii')
            username, password = auth_decoded.split(':', 2)
            # request.basicauth_user, request.basicauth_pass = username, password
        else:
            username = ''
            password = ''

        if self.application.settings['check_auth'](self.request, username, password):
            return

        self.set_status(401)
        self.set_header('WWW-Authenticate', 'Basic realm=' + self.application.settings['auth_realm'])
        self.write('401 Authentication Required')
        self.finish()



import collections
import logging


LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'


class LogFilter(object):
    def filter(self, record):
        if not record.name:
            return True
        if 'seesaw' in record.name or 'root' in record.name:
            return True


class InternalTempLogHandler(logging.Handler):
    """Logging handler that keeps recent lines to be retrieved later"""

    def __init__(self, *args, **kwargs):
        logging.Handler.__init__(self, *args, **kwargs)
        self.records = collections.deque((), 500)

    def emit(self, record):
        self.records.append(record)

    def get_str_list(self):
        return list(self.format(record) for record in self.records)



'''Managing work units.'''
import os
import os.path
import shutil
import traceback
import time
import collections

from seesaw.event import Event
import seesaw.six


class ItemData(collections.MutableMapping):
    '''Base item data property container.

    Args:
        properties (dict): Original dict
        on_property (Event): Fired whenever a property changes.
            Callback accepts:

            1. self
            2. key
            3. new value
            4. old value
    '''
    def __init__(self, properties=None):
        super(ItemData, self).__init__()
        self._properties = properties or {}
        self.on_property = Event()

    @property
    def properties(self):
        return self._properties

    def __getitem__(self, key):
        return self._properties[key]

    def __setitem__(self, key, value):
        old_value = self._properties.get(key, None)

        self._properties[key] = value

        if old_value != value:
            self.on_property(self, key, value, old_value)

    def __delitem__(self, key):
        old_value = self.properties.get(key, None)

        del self.properties[key]

        if old_value:
            self.on_property(self, key, None, old_value)

    def __len__(self):
        return len(self._properties)

    def __iter__(self):
        return iter(self._properties)


class Item(ItemData):
    '''A thing, or work unit, that needs to be downloaded.

    It has properties that are filled by the :class:`Task`.

    An Item behaves like a mutable mapping.

    .. note::
        State belonging to a item should be stored on the actual item
        itself. That is, do not store variables onto a :class:`Task` unless
        you know what you are doing.
    '''

    class ItemState(object):
        '''State of the item.'''
        running = "running"
        canceled = "canceled"
        completed = "completed"
        failed = "failed"

    class TaskStatus(object):
        '''Status of happened on a task.'''
        running = "running"
        completed = "completed"
        failed = "failed"

    def __init__(self, pipeline, item_id, item_number,
                 keep_data=False, prepare_data_directory=True,
                 **kwargs):
        super(Item, self).__init__(**kwargs)

        self._pipeline = pipeline
        self._item_id = item_id
        self._item_number = item_number
        self._keep_data = keep_data
        self.may_be_canceled = False

        self._item_state = self.ItemState.running
        self._task_status = {}
        self._start_time = time.time()
        self._end_time = None
        self._errors = []
        self._last_output = ""

        self.on_output = Event()
        self.on_error = Event()
        self.on_item_state = Event()
        self.on_task_status = Event()

        # Legacy events
        self.on_cancel = Event()
        self.on_complete = Event()
        self.on_fail = Event()
        self.on_finish = Event()
        self.on_item_state.handle(self._dispatch_legacy_events)

        if prepare_data_directory:
            self.prepare_data_directory()

    def __hash__(self):
        return hash(self._item_id)

    @property
    def item_state(self):
        return self._item_state

    @property
    def task_status(self):
        return self._task_status

    @property
    def start_time(self):
        return self._start_time

    @property
    def end_time(self):
        return self._end_time

    @property
    def canceled(self):
        return self._item_state == self.ItemState.canceled

    @property
    def completed(self):
        return self._item_state == self.ItemState.completed

    @property
    def failed(self):
        return self._item_state == self.ItemState.failed

    @property
    def finished(self):
        return self.canceled or self.completed or self.failed

    @property
    def pipeline(self):
        return self._pipeline

    @property
    def item_id(self):
        return self._item_id

    @property
    def item_number(self):
        return self._item_number

    def prepare_data_directory(self):
        dirname = os.path.join(self._pipeline.data_dir, self._item_id)
        self["data_dir"] = dirname
        if os.path.isdir(dirname):
            shutil.rmtree(dirname)
        os.makedirs(dirname)

    def clear_data_directory(self):
        if not self._keep_data:
            dirname = self["data_dir"]
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)

    def log_output(self, data, full_line=True):
        if isinstance(data, seesaw.six.binary_type):
            try:
                data = data.decode('utf8', 'replace')
            except UnicodeError:
                data = data.decode('ascii', 'replace')

        if full_line and len(data) > 0:
            if data[0] != "\n" and len(self._last_output) > 0 and \
                    self._last_output[-1] != "\n":
                data = "\n" + data
            if data[-1] != "\n":
                data += "\n"
        self._last_output = data
        self.on_output(self, data)

    def log_error(self, task, *args):
        self._errors.append((task, args))
        self.on_error(self, task, *args)

    def set_task_status(self, task, status):
        if task in self._task_status:
            old_status = self._task_status[task]
        else:
            old_status = None
        if status != old_status:
            self._task_status[task] = status
            self.on_task_status(self, task, status, old_status)

    def cancel(self):
        assert not self.canceled
        self.clear_data_directory()
        self._item_state = self.ItemState.canceled
        self._end_time = time.time()
        self.on_item_state(self, self._item_state)

    def complete(self):
        assert not self.completed
        self.clear_data_directory()
        self._item_state = self.ItemState.completed
        self._end_time = time.time()
        self.on_item_state(self, self._item_state)

    def fail(self):
        assert not self.failed
        self.clear_data_directory()
        self._item_state = self.ItemState.failed
        self._end_time = time.time()
        self.on_item_state(self, self._item_state)

    def _dispatch_legacy_events(self, item, state):
        if state == self.ItemState.failed:
            self.on_fail(item)
        elif state == self.ItemState.completed:
            self.on_complete(item)
        elif state == self.ItemState.canceled:
            self.on_cancel(item)
        else:
            raise Exception('Unknown event')

        self.on_finish(item)

    def description(self):
        return "Item %s" % self.properties.get("item_name", "")

    def __str__(self):
        return "<Item '{0}' {1} {2}>".format(
            self.properties.get("item_name", ""),
            self._item_id, self._item_state
        )
        # s = "Item " + ("FAILED " if self.failed else "") + str(self.properties)
        # for err in self._errors:
        #     for e in err[1]:
        #         # TODO this isn't how exceptions work?
        #         if isinstance(e, Exception):
        #             s += "%s\n" % traceback.format_exception(Exception, e,
        #                                                      None)
        #         else:
        #             s += "%s\n" % str(e)
        #     s += "\n  " + str(err)
        # return s


class ItemValue(object):
    '''Get an item's value during :func:`realize`.'''
    def __init__(self, key):
        self.key = key

    def realize(self, item):
        return item[self.key]

    def fill(self, item, value):
        if isinstance(self, ItemValue):
            item[self.key] = value
        elif self is None:
            pass
        else:
            raise Exception("Attempting to fill " + str(type(self)))

    def __str__(self):
        return "<" + self.key + ">"


class ItemInterpolation(object):
    '''Formats a string using the percent operator during :func:`realize`.'''
    def __init__(self, s):
        self.s = s

    def realize(self, item):
        return self.s % item

    def __str__(self):
        return "<'" + self.s + "'>"
